#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import numpy

import octobot_commons.databases.document_database_adaptors as adaptors
import octobot_commons.databases.bases.document_database as document_database
import octobot_commons.databases.database_caches as database_cache


class BaseDatabase:
    def __init__(
        self,
        file_path: str,
        database_adaptor=adaptors.TinyDBAdaptor,
        cache_size=None,
        enable_storage=True,
        **kwargs,
    ):
        self.enable_storage = enable_storage
        self._database = None
        if self.enable_storage and database_adaptor is not None:
            self._database = document_database.DocumentDatabase(
                database_adaptor(file_path, cache_size=cache_size, **kwargs)
            )
            self._database.initialize()
        self.are_data_initialized = False
        self.are_data_initialized_by_key = {}
        self.cache = database_cache.GenericDatabaseCache()

    def set_initialized_flags(self, value, keys=None):
        """
        Updates the initialized values of the given keys
        :param value: the updated initialized value
        :param keys: keys to update. Updated every key if left to None
        """
        self.are_data_initialized = value
        for key in keys or self.are_data_initialized_by_key.keys():
            self.are_data_initialized_by_key[key] = value

    def get_db_path(self):
        """
        :return: the path to the current database's path
        """
        return self._database.get_db_path()

    async def search(self, dict_query: dict = None):
        """
        :param dict_query: initialization dict for the query
        :return: a search query
        """
        if dict_query is None:
            return await self._database.query_factory()
        return (await self._database.query_factory()).fragment(dict_query)

    async def count(self, table_name: str, query) -> int:
        """
        :param table_name: the table to count data from
        :param query: the query to count results from
        :return: the number of elements that match the given query
        """
        return await self._database.count(table_name, query)

    async def flush(self):
        """
        Flushes the database, "committing" operations into the database
        """
        await self._database.flush()

    async def hard_reset(self):
        """
        Completely resets the database as if it just was created
        """
        return await self._database.hard_reset()

    def is_hard_reset_error(self, err: Exception) -> bool:
        """
        returns True if the given error should trigger
        a hard reset of the database
        """
        return self._database.is_hard_reset_error(err)

    async def close(self):
        """
        Closes the database, flushes it first
        """
        if self.enable_storage:
            await self.flush()
            await self._database.close()

    async def clear(self):
        """
        Clears the database, removing everything from it
        """
        self.cache.clear()

    async def contains_row(self, table: str, row: dict):
        """
        Returns true if the given rows are included in the given table, also looking into internal cache
        :param table: the table to look into
        :param row: the row to find
        """
        if self.cache.contains_row(table, row):
            return True
        return await self.count(table, await self.search(row)) > 0

    def __str__(self):
        return f"{self.__class__.__name__}, database: {self._database}"

    @classmethod
    def _create_database(
        cls,
        *args,
        required_adaptor=False,
        cache_size=None,
        database_adaptor=None,
        **kwargs,
    ):
        if required_adaptor:
            adaptor = kwargs.pop("database_adaptor", adaptors.TinyDBAdaptor)
            if adaptor is None:
                raise RuntimeError("database_adaptor parameter required")
            adaptor_instance = adaptor(*args, cache_size=cache_size, **kwargs)
            return (
                cls(
                    *args,
                    database_adaptor=database_adaptor,
                    cache_size=cache_size,
                    **kwargs,
                ),
                adaptor_instance,
            )
        return cls(*args, cache_size=cache_size, **kwargs), None

    @classmethod
    @contextlib.asynccontextmanager
    async def database(
        cls, *args, with_lock=False, cache_size=None, database_adaptor=None, **kwargs
    ):
        """
        Yields a database and closes it when exiting the context manager
        :param args: arguments to pass to the database constructor
        :param with_lock: When True, creating a lock synchronized database
        :param cache_size: size of the internal database cache
        :param database_adaptor: Database class to use
        :param kwargs: keyword arguments to pass to the database constructor
        """
        database, adaptor_instance = cls._create_database(
            *args,
            required_adaptor=with_lock,
            cache_size=cache_size,
            database_adaptor=database_adaptor,
            **kwargs,
        )
        if with_lock:
            async with document_database.DocumentDatabase.locked_database(
                adaptor_instance
            ) as locked_db:
                database._database = locked_db
                yield database
                # context manager is taking care of closing the database
            return
        try:
            yield database
        finally:
            await database.close()

    @staticmethod
    def get_serializable_value(value):
        """
        Returns a json serializable value of the given element. Mostly used to serialize numpy types
        :param value: the element
        """
        return value.item() if isinstance(value, numpy.generic) else value



# pylint: disable=W0718
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json
import os.path
import shutil
import jsonschema
import octobot_commons.logging
import octobot_commons.constants


LOGGER_NAME = "json_util"


def validate(config, schema_file) -> None:
    """
    Validate a config file, raise upon validation error
    :param config: the config
    :param schema_file: the config schema
    :return: None
    """
    with open(schema_file) as json_schema:
        loaded_schema = json.load(json_schema)
    jsonschema.validate(instance=config, schema=loaded_schema)


def has_same_content(file_path: str, expected_content: dict) -> bool:
    """
    :return: True if the content of the parsed json file at file_path equals the given expected_content
    """
    if os.path.isfile(file_path):
        content = read_file(file_path, raise_errors=False)
        return content == expected_content
    return False


def read_file(
    file_path: str,
    raise_errors: bool = True,
    on_error_value: dict = None,
    open_mode="r",
) -> dict:
    """
    Read a load the given file with json.load()
    :param file_path: file to read
    :param raise_errors: when True will forward errors. Will just log errors otherwise
    :param on_error_value: return this value when raise_errors is False and an error occurs
    :param open_mode: the file open mode to give to open()
    :return: the parsed file or default value on error if possible
    """
    try:
        with open(file_path, open_mode) as open_file:
            return json.load(open_file)
    except PermissionError as err:
        if raise_errors:
            raise
        octobot_commons.logging.get_logger(LOGGER_NAME).error(
            f"Permission error when reading {file_path} file: {err}."
        )
    except Exception as err:
        if raise_errors:
            raise
        octobot_commons.logging.get_logger(LOGGER_NAME).exception(
            f"Unexpected error when reading {file_path} file: {err}."
        )
    if on_error_value is None:
        raise ValueError("on_error_value is unset")
    return on_error_value


def safe_dump(content: dict, save_path: str, restore_file: str = None) -> None:
    """
    Safely dump content into save_path restoring the previous content if writing fails
    """
    restore_file = (
        restore_file or f"{save_path}{octobot_commons.constants.SAFE_DUMP_SUFFIX}"
    )
    try:
        has_initial_content = os.path.exists(save_path)
        if has_initial_content:
            if os.path.exists(restore_file):
                os.remove(restore_file)
            # prepare a restoration file
            shutil.copy(save_path, restore_file)
    except Exception as err:
        # when failing to create restore file
        error_details = (
            f"Failed to create the {restore_file} backup file. Is the associated folder  "
            f"folder accessible ? : {err} ({err.__class__.__name__})"
        )
        octobot_commons.logging.get_logger(LOGGER_NAME).exception(
            err, True, error_details
        )
        raise err.__class__(error_details) from err
    try:
        # create config content as str before opening file not to clear it on json dump exception
        str_content = dump_formatted_json(content)
        with open(save_path, "w") as write_file:
            write_file.write(str_content)

    except Exception as global_exception:
        # when failing to save the new file config
        octobot_commons.logging.get_logger(LOGGER_NAME).error(
            f"File save failed : {global_exception}. "
            f"{'restoring previous value' if has_initial_content else 'no previous value to restore'}"
        )
        if has_initial_content:
            # restore file with previous content
            shutil.copy(restore_file, save_path)
        elif os.path.exists(save_path):
            # no previous content: ensure no potentially created file is left
            os.remove(save_path)
        raise global_exception
    finally:
        # remove temporary restore file if any
        try:
            if os.path.exists(restore_file):
                os.remove(restore_file)
        except Exception as err:
            octobot_commons.logging.get_logger(LOGGER_NAME).exception(
                err, True, f"Failed to remove {restore_file} restore file: {err}"
            )


def dump_formatted_json(json_data) -> str:
    """
    The dumped json data
    :param json_data: the json data to be dumped
    :return: the dumped json data
    """
    return json.dumps(json_data, indent=4, sort_keys=True)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import octobot_commons.dataclasses.flexible_dataclass as flexible_dataclass


class MinimizableDataclass(flexible_dataclass.FlexibleDataclass):
    def to_dict(self, include_default_values=True) -> dict:
        """
        Creates a new dict from self. Recursively processes any MinimizableDataclass instance attribute
        """
        if include_default_values:
            # use default factory
            return dataclasses.asdict(self)
        factory = _asdict_without_default_factory(
            (self.__class__,)
            + tuple(
                getattr(self, attr.name)[0].__class__
                if isinstance(getattr(self, attr.name), list)
                and getattr(self, attr.name)
                else getattr(self, attr.name).__class__
                for attr in dataclasses.fields(self)
            )
        )
        return dataclasses.asdict(self, dict_factory=factory)


def _asdict_without_default_factory(possible_classes):
    def factory(obj) -> dict:
        formatted_dict = {}
        found_class = None
        for possible_class in possible_classes:
            if possible_class in (int, float, str, list, dict):
                continue
            if all(key in possible_class.__dataclass_fields__ for key, _ in obj):
                found_class = possible_class
        if found_class is None:
            # class not found, include all values
            return dict(obj)
        for key, val in obj:
            default_field_value = found_class.__dataclass_fields__[key].default
            if default_field_value is dataclasses.MISSING and (
                found_class.__dataclass_fields__[key].default_factory
                is not dataclasses.MISSING
            ):
                # try with default factory
                default_field_value = found_class.__dataclass_fields__[
                    key
                ].default_factory()
            if default_field_value is dataclasses.MISSING or default_field_value != val:
                formatted_dict[key] = val

        return formatted_dict

    return factory



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import traceback
import concurrent.futures

import octobot_commons.constants as constants
import octobot_commons.logging as logging_util


def run_coroutine_in_asyncio_loop(
    coroutine, async_loop, log_exceptions=True, timeout=constants.DEFAULT_FUTURE_TIMEOUT
):
    """
    Run a coroutine in the specified asyncio loop
    :param coroutine: the coroutine to run
    :param async_loop: the asyncio loop
    :param log_exceptions: logs exceptions when True
    :param timeout: number of seconds to wait for the future before raising a asyncio.TimeoutError
    :return: the execution result
    """
    logger = logging_util.get_logger("asyncio_tools")
    current_task_before_start = asyncio.current_task(async_loop)
    future = asyncio.run_coroutine_threadsafe(coroutine, async_loop)
    try:
        return future.result(timeout)
    except (asyncio.TimeoutError, concurrent.futures.TimeoutError) as timeout_error:
        logger.error(
            f"{coroutine} coroutine took too long to execute, cancelling the task. "
            f"(current task before starting this one: {current_task_before_start}, actual current "
            f"task before cancel: {asyncio.current_task(async_loop)})"
        )
        future.cancel()
        raise asyncio.TimeoutError from timeout_error
    except Exception as global_exception:
        if log_exceptions:
            logger.exception(
                global_exception,
                True,
                f"{coroutine} coroutine raised an exception: {global_exception}",
            )
        raise global_exception


class ErrorContainer:
    """
    ErrorContainer is used to catch exceptions in as asyncio loop context
    """

    def __init__(self):
        self.errors = []
        # set to True when investigating post loop closing exceptions (ex: on task __del__)
        self.print_received_exceptions = True

    def exception_handler(self, _, context) -> None:
        """
        To be set in the watched asyncio loop via loop.set_exception_handler()
        :param _: the loop argument, not used
        :param context: the context dict of the exception
        :return: None
        """
        self.errors.append(context)
        if self.print_received_exceptions:
            print(context)
            error = context.get("exception")
            if error:
                traceback.print_exception(
                    type(error), value=error, tb=error.__traceback__
                )

    async def check(self) -> None:
        """
        Will raise AssertionError if an exception has been raised in the registered loop(s)
        :return: None
        """
        if self.errors:
            raise AssertionError("\n".join(f"{e}" for e in self.errors))


async def wait_asyncio_next_cycle():
    """
    Wait for next asyncio next loop cycle
    """

    async def do_nothing():
        pass

    await asyncio.create_task(do_nothing())


class RLock(asyncio.Lock):
    """
    Async Lock implementing reentrancy
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task = None
        self._depth = 0

    async def acquire(self):
        if self._task is None or self._task is not asyncio.current_task():
            await super().acquire()
            self._task = asyncio.current_task()
            if self._depth != 0:
                raise RuntimeError(
                    f"Async RLock acquired when depth !=0 (depth = {self._depth})."
                )
        self._depth += 1
        return True

    def release(self):
        if self._depth > 0:
            self._depth -= 1
        if self._depth == 0:
            super().release()
            self._task = None



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.constants as commons_constants
import octobot_commons.singleton as singleton
import octobot_commons.logging as logging
import octobot_commons.async_job as async_job
import octobot_commons.authentication as authentication
import octobot_commons.profiles.profile_sharing as profile_sharing


class ProfileSynchronizer(singleton.Singleton):
    """
    Async job to maintain the profile associated to the given configuration up-to-date
    """

    DEFAULT_SYNC_REFRESH_INTERVAL = (
        commons_constants.PROFILE_REFRESH_HOURS_INTERVAL
        * commons_constants.HOURS_TO_SECONDS
    )

    def __init__(self, current_config, on_profile_change):
        super().__init__()
        self.current_config = current_config
        self._on_profile_change = on_profile_change
        self.sync_job = None
        self.sync_interval = self.DEFAULT_SYNC_REFRESH_INTERVAL
        self.logger = logging.get_logger(self.__class__.__name__)

    async def _sync_profile(self):
        if not self.current_config.profile.auto_update:
            self.logger.debug("Skipping profile update check: auto_update is False")
            return
        if not self.current_config.profile.slug:
            self.logger.error(
                "Impossible to check profile updates: profile slug is unset"
            )
            return
        self.logger.info(f"Synchronizing {self.current_config.profile.name} profile")
        if await profile_sharing.update_profile(self.current_config.profile):
            self.logger.info(f"{self.current_config.profile.name} profile updated")
            await self._on_profile_change(self.current_config.profile.name)
        else:
            self.logger.info(
                f"{self.current_config.profile.name} profile already up-to-date"
            )

    async def _should_sync_profiles(self):
        authenticator = authentication.Authenticator.instance()
        await authenticator.wait_for_private_data_fetch_if_processing()
        return authenticator.has_open_source_package()

    async def start(self) -> bool:
        """
        Synch the profile if necessary
        """
        if not await self._should_sync_profiles():
            self.logger.debug("Profile synch loop disabled")
            return False
        self.logger.debug("Starting profile synchronizer")
        self.sync_job = async_job.AsyncJob(
            self._sync_profile,
            first_execution_delay=0,
            execution_interval_delay=self.sync_interval,
        )
        await self.sync_job.run()
        return True

    def stop(self):
        """
        Stop the synchronization loop
        """
        if self.sync_job is not None and not self.sync_job.is_stopped():
            self.logger.debug("Stopping profile synchronizer")
            self.sync_job.stop()


async def start_profile_synchronizer(current_config, on_profile_change):
    """
    Start the clock synchronization loop if possible on this system
    :return: True if the loop has been started
    """
    return await ProfileSynchronizer.instance(current_config, on_profile_change).start()


async def stop_profile_synchronizer():
    """
    Stop the synchronization loop
    """
    return ProfileSynchronizer.instance().stop()
